function  model = team_training_code(input_directory,output_directory) % train_ECG_leads_classifier
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Purpose: Train ECG leads and obtain classifier models
% for 12-lead, 6-lead, 3-lead, 4-lead and 2-lead ECG sets
% Inputs:
% 1. input_directory
% 2. output_directory
%
% Outputs:
% model: trained model
% Logistic regression models for different sets of leads
%
% Author: Erick Andres Perez Alday, PhD, <perezald@ohsu.edu>
% Version 1.0 Aug-2020
% Revision History
% By: Nadi Sadr, PhD, <nadi.sadr@dbmi.emory.edu>
% Version 2.0 1-Dec-2020
% Version 2.2 25-Jan-2021
% Version 2.3 26-April-2021
%
% Reference: https://www.mathworks.com/help/wavelet/examples/classify-time-series-using-wavelet-analysis-and-deep-learning.html
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Define lead sets (e.g 12, 6, 4, 3 and 2 lead ECG sets)
twelve_leads = [{'I'}, {'II'}, {'III'}, {'aVR'}, {'aVL'}, {'aVF'}, {'V1'}, {'V2'}, {'V3'}, {'V4'}, {'V5'}, {'V6'}];
six_leads    = [{'I'}, {'II'}, {'III'}, {'aVR'}, {'aVL'}, {'aVF'}];
four_leads   = [{'I'}, {'II'}, {'III'}, {'V2'}];
three_leads  = [{'I'}, {'II'}, {'V2'}];
two_leads    = [{'I'}, {'II'}];
lead_sets = {twelve_leads, six_leads, four_leads, three_leads, two_leads};

disp('Loading data...')

% Find files.
input_files = {};
features =[];
for f = dir(input_directory)'
    if exist(fullfile(input_directory, f.name), 'file') == 2 && f.name(1) ~= '.' && all(f.name(end - 2 : end) == 'mat')
        input_files{end + 1} = f.name;
    end
end

% Extract classes from dataset.
% read number of unique classes
%classes = get_classes(input_directory,input_files);
classes = {'164889003',
    '164890007',
    '6374002',
    '426627000',
    '733534002',
    '713427006',
    '270492004',
    '713426002',
    '39732003',
    '445118002',
    '164909002',
    '251146004',
    '698252002',
    '426783006',
    '284470004',
    '10370003',
    '365413008',
    '427172004',
    '164947007',
    '111975006',
    '164917005',
    '47665007',
    '59118001',
    '427393009',
    '426177001',
    '427084000',
    '63593006',
    '164934002',
    '59931005',
    '17338001'};

num_classes = length(classes);     % number of classes
num_files   = length(input_files);
%Total_data  = cell(1,num_files);
%Total_header= cell(1,num_files);

%% Load data recordings and header files
% Iterate over files.
disp('Training model..')

label=zeros(num_files,num_classes);
NSR_only = zeros(1,num_files);
Single_Not_NSR = zeros(1,num_files);

for i = 1:num_files
    flagMS = 0;
    disp(['    ', num2str(i), '/', num2str(num_files), '...'])
    % Load data.
    file_tmp = strsplit(input_files{i},'.');
    tmp_input_file = fullfile(input_directory, file_tmp{1});
    [data,header_data] = load_challenge_data(tmp_input_file);
    
    %% Check the number of available ECG leads
    tmp_hea = strsplit(header_data{1},' ');
    num_leads = str2num(tmp_hea{2});
    [leads, leads_idx] = get_leads(header_data,num_leads);
    
    
    %% Extract labels
    for j = 1 : length(header_data)
        if startsWith(header_data{j},'#Dx')
            tmp = strsplit(header_data{j},': ');
            % Extract more than one label if available
            tmp_c = strsplit(tmp{2},',');
            if (length(tmp_c) == 1 && strcmp(tmp_c, '426783006'))
                NSR_only(i) = 1;
            end
            for k=1:length(tmp_c)
                if length(tmp_c) == 1
                    idx=find(strcmp(classes,tmp_c{k}));
                    if idx > 0
                        label(i,idx)=1;
                        flagMS = 1;
                        if ~strcmp(tmp_c, '426783006')
                            Single_Not_NSR(i) = 1;
                        end
                    end
                end
            end
            break
        end
    end
    
    if flagMS
        %% Create scalograms
        for k=1:num_leads
            sig = data(k,:);
            % get sampling frequency
            temp = strsplit(header_data{1});
            Fs = str2double(temp{3});
            fb = cwtfilterbank('SignalLength',length(sig),...
                'SamplingFrequency',Fs);
            [wt,~] = fb.wt(sig);
            saveScalogram(abs(wt),i,k);
        end
    end
    
end
%% create models
model = [];
filenames = cell(1,num_files);

for g=1:num_files
    filenames{g} = num2str(g);
    while length(filenames{g}) < 10
        filenames{g} = strcat('0',filenames{g});
    end
    
end

% create file list of examples with NSR only
% count how many NSR only samples there are in the dataset
numNSROnly = sum(NSR_only);
% find the indices of those samples
NSRidx = find(NSR_only == 1);

% compile a list of file numbers matching those samples
% correct it to match output format (10 characters)
filenamesNSR = cell(1,numNSROnly);
for m=1:length(NSRidx)
    filenamesNSR{m} = num2str(NSRidx(m));
    while length(filenamesNSR{m}) < 10
        filenamesNSR{m} = strcat('0',filenamesNSR{m});
    end
end

% create file list of examples of single dx not NSR
numSingle_Not_NSR = sum(Single_Not_NSR);
% find the indices of those samples
Single_Not_NSRidx = find(Single_Not_NSR == 1);

filenamesSingle_Not_NSR = cell(1,numSingle_Not_NSR);
for m=1:length(Single_Not_NSRidx)
    filenamesSingle_Not_NSR{m} = num2str(Single_Not_NSRidx(m));
    while length(filenamesSingle_Not_NSR{m}) < 10
        filenamesSingle_Not_NSR{m} = strcat('0',filenamesSingle_Not_NSR{m});
    end
end


% create a model for each scored arr
for kk=1:num_classes
    
    % find number of positive examples of this arrhythmia
    posSum = sum(label(:,kk));
    
    if posSum >= 1
        if ~strcmp(classes(kk),'426783006') % not NSR  --- NEW CHANGE
            % find indices of positive examples of this arrhythmia
            arrIdx = find(label(:,kk));
            images2Test = cell(1,2*posSum);
            labelTemp = zeros(1,2*posSum);
            % create cell array of filenames of positive and negative
            % examples of arrhythmia, alternating
            % get balanced number of negative examples
            count = 1;
            
            for h=1:2:2*posSum
                images2Test{1,h} = filenames{arrIdx(count)};
                labelTemp(h) = 1;
                loc = mod(count, numNSROnly)+1;
                % find matching number of positive examples of NSR
                images2Test{1,h+1} = filenamesNSR{1,loc};
                labelTemp(h+1) = 0;
                count = count + 1;
            end
        elseif strcmp(classes(kk),'426783006') % NSR
            % find indices of positive examples of this arrhythmia
            arrIdx = find(label(:,kk));
            images2Test = cell(1,2*posSum);
            labelTemp = zeros(1,2*posSum);
            % create cell array of filenames of positive and negative
            % examples of arrhythmia, alternating
            % get balanced number of negative examples
            count = 1;
            
            for h=1:2:2*posSum
                images2Test{1,h} = filenames{arrIdx(count)};
                labelTemp(h) = 1;
                loc = mod(count, numSingle_Not_NSR)+1;
                images2Test{1,h+1} = filenamesSingle_Not_NSR{1,loc};
                labelTemp(h+1) = 0;
                count = count + 1;
            end
            
            %images2Test = '*';
            %labelTemp = label(:,kk);
        end
        
        
        % separate model for each lead (per arr)
        for jj=1:12
           if (jj == 1 || jj == 2 || jj == 8)
            prefixNum = num2str(jj);
            allImages = imageDatastore(strcat(prefixNum, '_', images2Test, '.jpg'));
            allImages.Labels = categorical(labelTemp);
            
            sqz = squeezenet;
            lgraphSqz = layerGraph(sqz);
            tmpLayer = lgraphSqz.Layers(end-5);
            newDropoutLayer = dropoutLayer(0.6,'Name','new_dropout');
            lgraphSqz = replaceLayer(lgraphSqz,tmpLayer.Name,newDropoutLayer);
            
            numClasses = numel(categories(allImages.Labels));
            tmpLayer = lgraphSqz.Layers(end-4);
            newLearnableLayer = convolution2dLayer(1,numClasses, ...
                'Name','new_conv', ...
                'WeightLearnRateFactor',10, ...
                'BiasLearnRateFactor',10);
            lgraphSqz = replaceLayer(lgraphSqz,tmpLayer.Name,newLearnableLayer);
            tmpLayer = lgraphSqz.Layers(end);
            newClassLayer = classificationLayer('Name','new_classoutput');
            lgraphSqz = replaceLayer(lgraphSqz,tmpLayer.Name,newClassLayer);
            
            ilr = 3e-4;
            miniBatchSize = 10;
            maxEpochs = 15;
            valFreq = floor(numel(allImages.Files)/miniBatchSize);
            opts = trainingOptions('sgdm',...
                'MiniBatchSize',miniBatchSize,...
                'MaxEpochs',maxEpochs,...
                'InitialLearnRate',ilr,...
                'ExecutionEnvironment','gpu');
            
            model{kk,jj} = trainNetwork(allImages,lgraphSqz,opts);
            end
        end
    else
        for gg=1:12
            model{kk,gg} = {};
        end
    end
    clear arrIdx;
    clear images2Test;
    clear labelTemp;
end


%% save model data
for i=1:length(lead_sets)
    modelOut = [];
    num_leads = length(lead_sets{i});
    if (num_leads == 12)
        modelOut = model;
    elseif (num_leads == 6)
        modelOut = model(:, 1:6);
    elseif (num_leads == 4)
        modelOut = [model(:, 1:3), model(:,8)];
    elseif (num_leads == 3)
        modelOut = [model(:, 1:2), model(:,8)];
    elseif (num_leads == 2)
        modelOut = model(:, 1:2);
    end
    save_ECGleads_model(modelOut,output_directory,classes,num_leads);
    clear modelOut;
end
end

function save_ECGleads_model(model,output_directory,classes,num_leads) %save_ECG_model
% Save results.
tmp_file = [num2str(num_leads),'_lead_ecg_model.mat'];
filename = fullfile(output_directory,tmp_file);
save(filename,'model','classes','-v7.3');

disp('Done.')
end

function save_ECGleads_features(features,output_directory) %save_ECG_model
% Save results.
tmp_file = 'features.mat';
filename=fullfile(output_directory,tmp_file);
save(filename,'features');
end

% find unique number of classes
function classes = get_classes(input_directory,files)

classes={};
num_files = length(files);
k=1;
for i = 1:num_files
    g = strrep(files{i},'.mat','.hea');
    input_file = fullfile(input_directory, g);
    fid=fopen(input_file);
    tline = fgetl(fid);
    tlines = cell(0,1);
    
    while ischar(tline)
        tlines{end+1,1} = tline;
        tline = fgetl(fid);
        if startsWith(tline,'#Dx')
            tmp = strsplit(tline,': ');
            tmp_c = strsplit(tmp{2},',');
            for j=1:length(tmp_c)
                idx2 = find(strcmp(classes,tmp_c{j}));
                if isempty(idx2)
                    classes{k}=tmp_c{j};
                    k=k+1;
                end
            end
            break
        end
    end
    
    fclose(fid);
    
end
classes=sort(classes);
end

function [data,tlines] = load_challenge_data(filename)

% Opening header file
fid=fopen([filename '.hea']);

if (fid<=0)
    disp(['error in opening file ' filename]);
end

tline = fgetl(fid);
tlines = cell(0,1);
while ischar(tline)
    tlines{end+1,1} = tline;
    tline = fgetl(fid);
end
fclose(fid);

f=load([filename '.mat']);

try
    data = f.val;
    
    % take only 1st 5 seconds of data (shouldn't be less but check just in
    % case)
    temp = tlines{1};
    temp = strsplit(temp,' ');
    fs = str2num(cell2mat(temp(3)));
    if length(data) > fs*5
        data = data(:, 1:fs*5);
    end
catch ex
    rethrow(ex);
end

end
